from abc import ABC, abstractmethod
from math import floor

import numpy as np

from model_server.image import GenericImageFileAccessor


class Model(ABC):

    def __init__(self, autoload=True):
        """
        Abstract base class for an inference model that uses image data as an input.

        :param autoload: automatically load model and dependencies into memory if True
        """
        self.autoload = autoload
        if self.load():
            self.loaded = True
        else:
            self.loaded = False
            raise CouldNotLoadModelError()

    @abstractmethod
    def load(self):
        """
        Abstract method that carries out the expectedly time-consuming step of loading a model into memory
        :return: True if successful, else False
        """
        pass

    @abstractmethod
    def infer(self,
              img: GenericImageFileAccessor,
              channel: int = None
              ) -> (np.ndarray, dict): # return json describing inference result

        if self.autoload:
            self.load()

        if channel and channel >= img.chroma:
            raise ChannelTooHighError(f'Requested channel {channel} but image contains only {img.chroma} channels')

    def reload(self):
        self.load()

class ImageToImageModel(Model):

    def __init__(self, **kwargs):
        """
        Abstract class for models that receive an image and return an image of the same size
        :param kwargs: variable length keyword arguments
        """
        return super(**kwargs)

    @abstractmethod
    def infer(self, img, channel=None) -> (np.ndarray, dict):
        super().infer(img, channel)

class IlastikImageToImageModel(ImageToImageModel):
    pass

class DummyImageToImageModel(Model):

    model_id = 'dummy_make_white_square'

    def load(self):
        return True

    def infer(self, img: GenericImageFileAccessor, channel=None) -> (np.ndarray, dict):
        super().infer(img, channel)
        w = img.shape_dict['X']
        h = img.shape_dict['Y']
        result = np.zeros([h, w], dtype='uint8')
        result[floor(0.25 * h) : floor(0.75 * h), floor(0.25 * w) : floor(0.75 * w)] = 255
        return (result, {'success': True})


class Error(Exception):
    pass

class ChannelTooHighError(Error):
    pass

class CouldNotLoadModelError(Error):
    pass